local ADDON_NAME, NAMESPACE = ...;

local Addon = CreateFrame('Frame');
Addon:SetScript('OnEvent', function(self, event, ...) return self[event](self, ...); end);

NAMESPACE[1] = Addon;
NAMESPACE[2] = {};

_G.Emoji = NAMESPACE[1];

local EmojiPacks      = {};
local EmojiTextureMap = {};

local INCOMPLETE_SHORTCODE = ':[^:][^:]+$';
local COMPLETE_SHORTCODE   = ':[^:]+:';

local CHAT_MSG_TYPES = {
    'AFK',
    'BATTLEGROUND_LEADER',
    'BATTLEGROUND',
    'CHANNEL',
    'DND',
    'EMOTE',
    'GUILD',
    'OFFICER',
    'PARTY_LEADER',
    'PARTY',
    'RAID_LEADER',
    'RAID_WARNING',
    'RAID',
    'SAY',
    'WHISPER',
    'WHISPER_INFORM',
    'BN_WHISPER',
    'BN_WHISPER_INFORM',
    'YELL'
};

Addon.SetPack = function(name, path, strategy)
    EmojiPacks[name] = {
        path     = path,
        strategy = strategy,
    };

    -- Regenerate EmojiTextureMap
    EmojiTextureMap = {}
    for packName, pack in pairs(EmojiPacks) do
        for _, emoji in pairs(pack.strategy) do
            for _, emojiName in pairs(emoji.names) do
                EmojiTextureMap[emojiName] = '\124T' .. pack.path .. '\\' .. emoji.file;
            end
        end
    end
end

Addon.GetPacks = function()
    return EmojiPacks;
end

Addon.GetPacksCount = function()
    local count = 0;

    for _ in pairs(EmojiPacks) do
        count = count + 1;
    end

    return count;
end

local function startsWith(str, start)
    return str:sub(1, #start) == start;
end

local function oneStartsWith(table, start)
    for _, v in pairs(table) do
        if startsWith(v, start) then
            return true;
        end
    end

    return false;
end

local results = {};
local function getMatches(text, numResults)
    table.wipe(results);

    for name, pack in pairs(EmojiPacks) do
        for _, emoji in pairs(pack.strategy) do
            if oneStartsWith(emoji.names, text) then
                table.insert(
                    results,
                    {
                        ['name']     = '\124T' .. pack.path .. '\\' .. emoji.file .. ':0\124t ' .. emoji.names[1],
                        ['priority'] = LE_AUTOCOMPLETE_PRIORITY_OTHER
                    }
                );
            end

            if #results >= numResults then
                return results;
            end
        end
    end

    return results;
end

local function completeFunction(editBox, _, nameInfo)
    local complete = nameInfo.name:match(COMPLETE_SHORTCODE .. '$');
    local cursorPosition = editBox:GetCursorPosition();
    local text = editBox:GetText();
    local beginning = text:sub(1, cursorPosition);
    local incomplete = beginning:match(INCOMPLETE_SHORTCODE);

    if complete and incomplete then
        if complete and incomplete then
            local newText = beginning:sub(1, -(#incomplete) - 1) .. complete .. ' ';
            editBox:SetText(newText .. text:sub(cursorPosition + 1));
            editBox:SetCursorPosition(#newText);
        end

        return true;
    end

    return false;
end

local function onTextChanged(self, userInput)
    if userInput == false then
        return;
    end

    local text = self:GetText();
    local beginning = text:sub(1, self:GetCursorPosition());
    local shortCode = beginning:match(INCOMPLETE_SHORTCODE);

    if shortCode then
        AutoCompleteEditBox_SetAutoCompleteSource(self, getMatches);
        AutoCompleteEditBox_SetCustomAutoCompleteFunction(self, completeFunction);
        AutoComplete_Update(self, shortCode, #shortCode);
    end
end

local links = {};
local function onChatMessage(self, event, message, ...)
    table.wipe(links);

    local plusSize = 0;
    if _G.SELECTED_DOCK_FRAME then
        plusSize = select(2, _G.SELECTED_DOCK_FRAME:GetFont()) + 8;
    end

    message =
        message:gsub(
        '\124H.-\124h.-\124h',
        function(link)
            table.insert(links, link);
            return ('\2' .. #links .. '\3');
        end
    )

    message =
        message:gsub(
        COMPLETE_SHORTCODE,
        function(shortCode)
            local texture = EmojiTextureMap[shortCode];
            if texture then
                return texture .. ':' .. plusSize .. '|t';
            else
                return shortCode;
            end
        end
    )

    if #links > 0 then
        for index, link in pairs(links) do
            message = message:gsub('\2' .. index .. '\3', link);
        end
    end

    return false, message, ...;
end

Addon:RegisterEvent('ADDON_LOADED');
function Addon:ADDON_LOADED(addonName)
    if addonName ~= ADDON_NAME then
        return;
    end

    EmojiDB = EmojiDB or {};
    EmojiDB.recent = EmojiDB.recent or {};

    for i = 1, NUM_CHAT_WINDOWS do
        local editbox = _G['ChatFrame' .. i .. 'EditBox'];
        if editbox then
            editbox:SetScript('OnTextChanged', onTextChanged);
            editbox.HasStickyFocus = function() return true; end
        end
    end

    for _, t in pairs(CHAT_MSG_TYPES) do
       ChatFrame_AddMessageEventFilter('CHAT_MSG_' .. t, onChatMessage);
    end

    hooksecurefunc('ChatEdit_UpdateHeader', function(editBox)
        local chatType = editBox:GetAttribute('chatType');
        if not chatType then
            return;
        end

        local header = _G[editBox:GetName()..'Header'];
        local headerSuffix = _G[editBox:GetName()..'HeaderSuffix'];

        if not header then
            return;
        end

        editBox:SetTextInsets(15 + header:GetWidth() + (headerSuffix:IsShown() and headerSuffix:GetWidth() or 0), 13 + 20, 0, 0);
    end);

    self:UnregisterEvent('ADDON_LOADED');
end